!********************************************************************************
!** CONSTANT-NPT MOLECULAR DYNAMICS - EXTENDED SYSTEM METHOD                   **
!   Parrinello-Rahman / Nose methods
!   (Modified by Yoshi, Jan 2005, from Allen-Tildesley's F28)
!********************************************************************************

!C    *******************************************************************
!C    ** CONSTANT-NVT MOLECULAR DYNAMICS USING AN EXTENDED SYSTEM.     **
!C    **                                                               **
!C    ** REFERENCE:                                                    **
!C    **                                                               **
!C    ** NOSE, MOLEC. PHYS. 52, 255, 1984.                             **
!C    **                                                               **
!C    ** ROUTINES:                                                     **
!C    **                                                               **
!C    ** SUBROUTINE PREDIC                                             **
!C    **    PREDICTS POSITIONS AT TIME T + DT                          **
!C    ** SUBROUTINE CORREC                                             **
!C    **    CORRECTS POSITIONS AT TIME T + DT                          **
!C    **                                                               **
!C    ** PRINCIPAL VARIABLES:                                          **
!C    **                                                               **
!C    ** INTEGER N                             NUMBER OF ATOMS         **
!C    ** REAL    DT                            TIMESTEP                **
!C    ** REAL    RX(N),RY(N),RZ(N)             ATOM POSITIONS          **
!C    ** REAL    VX(N),VY(N),VZ(N)             FIRST DERIVATIVES       **
!C    ** REAL    AX(N),AY(N),AZ(N)             SECOND DERIVATIVES      **
!C    ** REAL    BX(N),BY(N),BZ(N)             THIRD DERIVATIVES       **
!C    ** REAL    FX(N),FY(N),FZ(N)             TOTAL FORCES            **
!C    ** REAL    V,VV,VA,VB,VC,VF              S AND DERIVATIVES       **
!C    ** REAL    WW                            MASS OF PISTON          **
!C    **                                                               **
!C    ** USAGE:                                                        **
!C    **                                                               **
!C    ** THE MODIFIED EQUATIONS OF MOTION ARE AS FOLLOWS:              **
!C    ** A  = F/WM + (R/3)[AV/V - (2/3)*(VV/V)**2] - SV*V/S            **
!C    ** SA = SUM (WM*S/WQ)*(V-VV/(3V)*R)**2  - 3N*KT*S/WQ +SV**2/S    **
!C    ** VA = ( PRES - PRESUR ) / WW * S ** 2 + SV/S*VV                **
!C    **   PRES = (1/3V) * [ WM * SUM{(V-R*(VV/3V))**2} + VIRIAL       **
!C    **   VIRIAL = SUM{R dot F}                                       **
!C    ** WHERE R,V,A,.. ARE SUCCESSIVE DERIVATIVES OF POSITION         **
!C    ** AND SV,SA,.... ARE SUCCESSIVE DERIVATIVES OF EXTRA VARIABLE S **
!C    ** AND VV,VA,.... ARE SUCCESSIVE DERIVATIVES OF VOLUME           **
!C    ** WE SOLVE THESE EQUATIONS BY A GEAR 5-VALUE METHOD FOR SECOND  **
!C    ** ORDER DIFFERENTIAL EQUATIONS.                                 **
!C    ** THIS PROGRAM USES UNIT 10 FOR CONFIGURATION INPUT AND OUTPUT  **
!C    *******************************************************************

!C    *******************************************************************
!C    ** STANDARD TAYLOR SERIES PREDICTORS                             **
!C    *******************************************************************

        SUBROUTINE gear_pc_nphpr_predic_rlx
        use variables
        use constants
        implicit double precision (a-h,o-z)

        INTEGER :: I
        double precision :: C1, C2, C3, C4

        double precision, parameter :: damper_param = 1.0d13

!C    *******************************************************************

        C1 = DT
        C2 = C1 * DT / 2.0
        C3 = C2 * DT / 3.0
        C4 = C3 * DT / 4.0

        DO 100 I = 1, N

           RX(I) = RX(I) + C1*VX(I) + C2*AX(I) + C3*BX(I) + C4*CX(I)
           RY(I) = RY(I) + C1*VY(I) + C2*AY(I) + C3*BY(I) + C4*CY(I)
           RZ(I) = RZ(I) + C1*VZ(I) + C2*AZ(I) + C3*BZ(I) + C4*CZ(I)
           VX(I) = VX(I) + C1*AX(I) + C2*BX(I) + C3*CX(I)
           VY(I) = VY(I) + C1*AY(I) + C2*BY(I) + C3*CY(I)
           VZ(I) = VZ(I) + C1*AZ(I) + C2*BZ(I) + C3*CZ(I)
           AX(I) = AX(I) + C1*BX(I) + C2*CX(I)
           AY(I) = AY(I) + C1*BY(I) + C2*CY(I)
           AZ(I) = AZ(I) + C1*BZ(I) + C2*CZ(I)
           BX(I) = BX(I) + C1*CX(I)
           BY(I) = BY(I) + C1*CY(I)
           BZ(I) = BZ(I) + C1*CZ(I)

100     CONTINUE

        HMAT  = HMAT  + C1*HVMAT + C2*HAMAT + C3*HBMAT + C4*HCMAT
        HVMAT = HVMAT + C1*HAMAT + C2*HBMAT + C3*HCMAT
        HAMAT = HAMAT + C1*HBMAT + C2*HCMAT + HVMAT * damper_param
        HBMAT = HBMAT + C1*HCMAT

        RETURN
        END

!C    *******************************************************************
!C    ** GEAR CORRECTOR ALGORITHM                                      **
!C    **                                                               **
!C    ** WE USE GEAR'S ALTERNATIVE COEFFICIENT GEAR0 WHICH APPLIES IN  **
!C    ** THE CASE THAT FIRST TIME DERIVATIVES APPEAR ON THE RIGHT OF   **
!C    ** THESE SECOND-ORDER DIFFERENTIAL EQUATIONS.                    **
!C    ** FOR TIMESTEP-SCALED VARIABLES THE COEFFICIENTS WOULD BE       **
!C    ** 19/90, 3/4, 1, 1/2, 1/12.                                     **
!C    **                                                               **
!C    ** REFERENCES:                                                   **
!C    **                                                               **
!C    ** GEAR, NUMERICAL INITIAL VALUE PROBLEMS IN ORDINARY            **
!C    ** DIFFERENTIAL EQUATIONS (PRENTICE-HALL, 1971).                 **
!C    ** GEAR, REPORT ANL 7126 (ARGONNE NATIONAL LAB., 1966).          **
!C    *******************************************************************

        SUBROUTINE gear_pc_nphpr_correc_rlx
        use variables
        use constants
        implicit double precision (a-h,o-z)

        double precision :: GEAR0, GEAR1, GEAR3, GEAR4
        PARAMETER ( GEAR0 = 19.0 / 90.0, GEAR1 = 3.0 /4.0, &
        GEAR3 = 1.0 / 2.0,   GEAR4 = 1.0 /12.0 )

        DOUBLE PRECISION :: GMAT(3,3),GVMAT(3,3),GINMAT(3,3)
        DOUBLE PRECISION :: PT(3,3) ! PRESSURE TENSOR
        DOUBLE PRECISION :: TM(3,3),TM2(3,3),TM3(3,3) ! MATRIX FOR TEMPORARY USE
        DOUBLE PRECISION :: HF(3,3) ! FOR CALCULATION OF CORRECTOR FOR HMAT
        DOUBLE PRECISION :: REC(3,3) ! RECIPROCAL CELL MATRIX
        DOUBLE PRECISION :: CORRM(3,3)

!C    *******************************************************************

!  Inverse cell matrix
        CALL INVERSE (HMAT,HINMAT)
!  CALC VOLUME
        CALL CALCVOLUME
!  RECIPROCAL CELL MATRIX
        CALL RECIPRO (HMAT,REC)

!  Kinetic energy has to be calculated using P. Not V.
!  PRESSURE TENSOR IS ALSO CALCULATED.
        enkin = 0.0D0; enkineach=0.0d0
        PT = 0.0D0
        TM = MATMUL(HVMAT,HINMAT)
        DO I = 1, N
           XX = VX(I) - ( TM(1,1) * RX(I) + TM(1,2) * RY(I) + TM(1,3) * RZ(I) )
           YY = VY(I) - ( TM(2,1) * RX(I) + TM(2,2) * RY(I) + TM(2,3) * RZ(I) )
           ZZ = VZ(I) - ( TM(3,1) * RX(I) + TM(3,2) * RY(I) + TM(3,3) * RZ(I) )
           enkineach(i) = 0.5d0 * wm(i) * (XX ** 2 + YY ** 2 + ZZ ** 2)
           enkin = enkin + enkineach(i)
           PT(1,1) = PT(1,1) + XX * XX * WM(I)
           PT(2,1) = PT(2,1) + YY * XX * WM(I)
           PT(3,1) = PT(3,1) + ZZ * XX * WM(I)
           PT(1,2) = PT(1,2) + XX * YY * WM(I)
           PT(2,2) = PT(2,2) + YY * YY * WM(I)
           PT(3,2) = PT(3,2) + ZZ * YY * WM(I)
           PT(1,3) = PT(1,3) + XX * ZZ * WM(I)
           PT(2,3) = PT(2,3) + YY * ZZ * WM(I)
           PT(3,3) = PT(3,3) + ZZ * ZZ * WM(I)
        ENDDO
        PT = PT + DMAT
        PT = PT / V

        HF = MATMUL((PT - PEXMAT), REC) / WW

        C1 = DT
        C2 = C1 * DT / 2.0
        C3 = C2 * DT / 3.0
        C4 = C3 * DT / 4.0

        CR = GEAR0 * C2
        CV = GEAR1 * C2 / C1
        CB = GEAR3 * C2 / C3
        CC = GEAR4 * C2 / C4

        GMAT = MATMUL(TRANSPOSE(HMAT),HMAT)
        GVMAT = MATMUL(TRANSPOSE(HVMAT),HMAT) + MATMUL(TRANSPOSE(HMAT),HVMAT)
        TM2 = MATMUL(GVMAT,HINMAT)
        TM2 = MATMUL(TRANSPOSE(HINMAT),TM2)
        TM2 = MATMUL(HINMAT,TM2)
        TM2 = MATMUL(HMAT,TM2)
        TM3 = MATMUL(HAMAT,HINMAT)
        
        DO 400 I = 1, N

           XX = VX(I) - ( TM(1,1) * RX(I) + TM(1,2) * RY(I) + TM(1,3) * RZ(I) )
           YY = VY(I) - ( TM(2,1) * RX(I) + TM(2,2) * RY(I) + TM(2,3) * RZ(I) )
           ZZ = VZ(I) - ( TM(3,1) * RX(I) + TM(3,2) * RY(I) + TM(3,3) * RZ(I) )
           AXI = FX(I) / WM(I)    -  (TM2(1,1)*XX+TM2(1,2)*YY+TM2(1,3)*ZZ) &
           + 2.0D0*(TM(1,1) *XX+TM(1,2) *YY+TM(1,3) *ZZ) &
           + (TM3(1,1)*RX(I)+TM3(1,2)*RY(I)+TM3(1,3)*RZ(i))
           AYI = FY(I) / WM(I)    -  (TM2(2,1)*XX+TM2(2,2)*YY+TM2(2,3)*ZZ) &
           + 2.0D0*(TM(2,1) *XX+TM(2,2) *YY+TM(2,3) *ZZ) &
           + (TM3(2,1)*RX(I)+TM3(2,2)*RY(I)+TM3(2,3)*RZ(i))
           AZI = FZ(I) / WM(I)    -  (TM2(3,1)*XX+TM2(3,2)*YY+TM2(3,3)*ZZ) &
           + 2.0D0*(TM(3,1) *XX+TM(3,2) *YY+TM(3,3) *ZZ) &
           + (TM3(3,1)*RX(I)+TM3(3,2)*RY(I)+TM3(3,3)*RZ(i))

           CORRX = AXI - AX(I)
           CORRY = AYI - AY(I)
           CORRZ = AZI - AZ(I)

           RX(I) = RX(I) + CR * CORRX
           RY(I) = RY(I) + CR * CORRY
           RZ(I) = RZ(I) + CR * CORRZ
           VX(I) = VX(I) + CV * CORRX
           VY(I) = VY(I) + CV * CORRY
           VZ(I) = VZ(I) + CV * CORRZ
           AX(I) = AXI
           AY(I) = AYI
           AZ(I) = AZI
           BX(I) = BX(I) + CB * CORRX
           BY(I) = BY(I) + CB * CORRY
           BZ(I) = BZ(I) + CB * CORRZ
           CX(I) = CX(I) + CC * CORRX
           CY(I) = CY(I) + CC * CORRY
           CZ(I) = CZ(I) + CC * CORRZ

400     CONTINUE

        CORRM = HF - HAMAT
        HMAT  = HMAT  + CR * CORRM
        HVMAT = HVMAT + CV * CORRM
        HAMAT = HF
        HBMAT = HBMAT + CB * CORRM
        HCMAT = HCMAT + CC * CORRM

        CALL CALCVOLUME

        RETURN
        END

